/*
 * Copyright (c) 2010, Marco Brade
 * 							[https://sourceforge.net/users/mbrade] All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright notice,
 *       this list of conditions and the following disclaimer in the documentation
 *       and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package net.sf.prefixedproperties;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.Enumeration;
import java.util.ResourceBundle;

/**
 * The Class PrefixedPropertyResourceBundle.
 */
public class PrefixedPropertyResourceBundle extends ResourceBundle {

    private final PrefixedProperties properties;

    /**
     * Instantiates a new prefixed property resource bundle.
     * 
     * @param properties
     *            the properties
     */
    public PrefixedPropertyResourceBundle(final PrefixedProperties properties) {
	this.properties = properties;
    }

    /**
     * Creates the prefixed property resource bundle from json. The given reader will stay open after reading.
     * 
     * @param reader
     *            the reader
     * @param defaultPrefix
     *            the default prefix
     * @return the prefixed property resource bundle
     * @throws IOException
     *             Signals that an I/O exception has occurred.
     */
    public static PrefixedPropertyResourceBundle createPrefixedPropertyResourceBundleFromJSON(final Reader reader, final String defaultPrefix) throws IOException {
	final PrefixedProperties properties = new PrefixedProperties(defaultPrefix);
	properties.loadFromJSON(reader);
	return new PrefixedPropertyResourceBundle(properties);
    }

    /**
     * Creates the prefixed property resource bundle from properties. The given reader will stay open after reading.
     * 
     * @param reader
     *            the reader
     * @param defaultPrefix
     *            the default prefix
     * @return the prefixed property resource bundle
     * @throws IOException
     *             Signals that an I/O exception has occurred.
     */
    public static PrefixedPropertyResourceBundle createPrefixedPropertyResourceBundleFromProperties(final Reader reader, final String defaultPrefix) throws IOException {
	final PrefixedProperties properties = new PrefixedProperties(defaultPrefix);
	properties.load(reader);
	return new PrefixedPropertyResourceBundle(properties);
    }

    /**
     * Creates the prefixed property resource bundle from xml. The given InputStream will stay open after reading.
     * 
     * @param is
     *            the is
     * @param defaultPrefix
     *            the default prefix
     * @return the prefixed property resource bundle
     * @throws IOException
     *             Signals that an I/O exception has occurred.
     */
    public static PrefixedPropertyResourceBundle createPrefixedPropertyResourceBundleFromXML(final InputStream is, final String defaultPrefix) throws IOException {
	final PrefixedProperties properties = new PrefixedProperties(defaultPrefix);
	properties.loadFromXML(is);
	return new PrefixedPropertyResourceBundle(properties);
    }

    /* (non-Javadoc)
     * @see java.util.ResourceBundle#getKeys()
     */
    @Override
    public Enumeration<String> getKeys() {
	return properties.propertyNames();
    }

    /**
     * Sets the configured prefic.
     * 
     * @param configuredPrefix
     *            the new configured prefic
     */
    public void setConfiguredPrefic(final String configuredPrefix) {
	properties.setConfiguredPrefix(configuredPrefix);
    }

    /**
     * Sets the default prefix.
     * 
     * @param prefix
     *            the new default prefix
     */
    public void setDefaultPrefix(final String prefix) {
	properties.setDefaultPrefix(prefix);
    }

    /* (non-Javadoc)
     * @see java.util.ResourceBundle#handleGetObject(java.lang.String)
     */
    @Override
    protected Object handleGetObject(final String key) {
	if (key == null) {
	    throw new NullPointerException("The given key is null.");
	}
	return properties.get(key);
    }
}
